"""
TEI (Text Encoding Initiative) XML processing module
"""

import datetime
import hashlib

from bs4 import BeautifulSoup
from dateutil import parser
from nltk.tokenize import sent_tokenize

from ..analysis import Study
from ..grammar import Grammar
from ..schema.article import Article
from ..table import Table
from ..text import Text

# Global helper for multi-processing support
# pylint: disable=W0603
GRAMMAR = None

def getGrammar():
    """
    Multiprocessing helper method. Gets (or first creates then gets) a global grammar object to
    be accessed in a new subprocess.

    Returns:
        Grammar
    """

    global GRAMMAR

    if not GRAMMAR:
        GRAMMAR = Grammar()

    return GRAMMAR

class TEI(object):
    """
    Methods to transform TEI (Text Encoding Initiative) XML into article objects.
    """

    @staticmethod
    def date(published):
        """
        Attempts to parse a publication date, if available. Otherwise, None is returned.

        Args:
            published: published object

        Returns:
            publication date if available/found, None otherwise
        """

        # Parse publication date
        # pylint: disable=W0702
        try:
            published = parser.parse(published["when"]) if published and "when" in published.attrs else None
        except:
            published = None

        return published

    @staticmethod
    def authors(source):
        """
        Builds an authors string from a TEI sourceDesc tag.

        Args:
            source: sourceDesc tag handle

        Returns:
            semicolon separated list of authors
        """

        authors = []
        for name in source.find_all("persname"):
            surname = name.find("surname")
            forename = name.find("forename")

            if surname and forename:
                authors.append("%s, %s" % (surname.text, forename.text))

        return "; ".join(authors)

    @staticmethod
    def metadata(soup):
        """
        Extracts article metadata.

        Args:
            soup: bs4 handle

        Returns:
            (published, publication, authors, reference)
        """

        # Build reference link
        source = soup.find("sourcedesc")
        if source:
            published = source.find("monogr").find("date")
            publication = source.find("monogr").find("title")

            # Parse publication information
            published = TEI.date(published)
            publication = publication.text if publication else None
            authors = TEI.authors(source)

            struct = soup.find("biblstruct")
            reference = "https://doi.org/" + struct.find("idno").text if struct and struct.find("idno") else None
        else:
            published, publication, authors, reference = None, None, None, None

        return (published, publication, authors, reference)

    @staticmethod
    def abstract(soup, title):
        """
        Builds a list of title and abstract sections.

        Args:
            soup: bs4 handle
            title: article title

        Returns:
            list of sections
        """

        sections = [("TITLE", title)]

        abstract = soup.find("abstract").text
        if abstract:
            # Transform and clean text
            abstract = Text.transform(abstract)
            abstract = abstract.replace("\n", " ")

            sections.extend([("ABSTRACT", x) for x in sent_tokenize(abstract)])

        return sections

    @staticmethod
    def text(soup, title):
        """
        Builds a list of text sections.

        Args:
            soup: bs4 handle
            title: article title

        Returns:
            list of sections
        """

        # Initialize with title and abstract text
        sections = TEI.abstract(soup, title)

        for section in soup.find("text").find_all("div", recursive=False):
            # Section name and text
            children = list(section.children)

            # Attempt to parse section header
            if not children[0].name:
                name = str(children[0]).upper()
                children = children[1:]
            else:
                name = None

            text = " ".join([str(e.text) for e in children])
            text = text.replace("\n", " ")

            # Transform and clean text
            text = Text.transform(text)

            # Split text into sentences, transform text and add to sections
            sections.extend([(name, x) for x in sent_tokenize(text)])

        # Extract text from tables
        for figure in soup.find("text").find_all("figure"):
            # Get figure name
            name = str(figure.contents[0])

            # Search for table
            table = figure.find("table")
            if table:
                sections.extend([(name, x) for x in Table.extract(table)])

        return sections

    @staticmethod
    def parse(stream, source, models):
        """
        Parses a TEI XML datastream and returns a processed article.

        Args:
            stream: handle to input data stream
            source: text string describing stream source, can be None
            models: path to study models

        Returns:
            Article
        """

        # Get grammar handle
        grammar = getGrammar()

        soup = BeautifulSoup(stream, "lxml")

        title = soup.title.text

        # Extract article metadata
        published, publication, authors, reference = TEI.metadata(soup)

        # Parse text sections
        sections = TEI.text(soup, title)

        # Build NLP tokens for sections
        tokenslist = grammar.parse([text for _, text in sections])

        # Join NLP tokens with sections
        sections = [(name, text, tokenslist[x]) for x, (name, text) in enumerate(sections) if tokenslist[x]]

        # Parse study design fields
        design, size, sample, method, labels = Study.parse(sections, models)

        # Add additional fields to each section
        sections = [(name, text, labels[x] if labels[x] else grammar.label(tokens)) for x, (name, text, tokens) in enumerate(sections)]

        # Derive uid
        uid = hashlib.sha1(title.encode("utf-8") if title else reference.encode("utf-8")).hexdigest()

        # Default title to source if empty
        title = title if title else source

        # Article metadata - id, source, published, publication, authors, title, tags, design, sample size
        #                    sample section, sample method, reference, entry date
        metadata = (uid, "PDF", published, publication, authors, title, "PDF", design, size,
                    sample, method, reference, datetime.datetime.now().strftime("%Y-%m-%d"))

        return Article(metadata, sections, source)



"""
Elasticsearch module
"""

from elasticsearch import Elasticsearch, helpers

from .database import Database

class Elastic(Database):
    """
    Defines data structures and methods to store article content in Elasticsearch.
    """

    # Articles index
    ARTICLES = {
        "settings" : {
            "number_of_shards" : 5,
            "number_of_replicas" : 0,
            "index.mapping.nested_objects.limit": 30000
        },
        "mappings": {
            "properties" : {
                "sections" : {"type" : "nested"}
            }
        }
    }

    def __init__(self, url):
        """
        Connects and initializes an elasticsearch instance.

        Args:
            url: elasticsearch url
        """

        # Connect to ES instance
        self.connection = Elasticsearch(hosts=[url], timeout=60, retry_on_timeout=True)

        # Row count
        self.rows = 0

        # Buffered actions
        self.buffer = []

        # Create index if it doesn't exist
        if not self.connection.indices.exists("articles"):
            self.connection.indices.create("articles", Elastic.ARTICLES)

    def save(self, article):
        # Build article
        article = article.build()

        # Bulk action fields
        article["_id"] = article["id"]
        article["_index"] = "articles"

        # Buffer article
        self.buffer.append(article)

        # Increment number of articles processed
        self.rows += 1

        # Bulk load every 1000 records
        if self.rows % 1000 == 0:
            helpers.bulk(self.connection, self.buffer)
            self.buffer = []

            print("Inserted {} articles".format(self.rows), end="\r")

    def complete(self):
        # Load remaining buffered articles
        if self.buffer:
            helpers.bulk(self.connection, self.buffer)

        print("Total articles inserted: {}".format(self.rows))

        # Refresh indices
        self.connection.indices.refresh(index="articles")

    def close(self):
        self.connection.close()



"""
Design module

Labels definitions:

  1 - Systematic review
  2 - Randomized control trial
  3 - Non-randomized trial
  4 - Prospective observational
  5 - Time-to-event analysis
  6 - Retrospective observational
  7 - Cross-sectional
  8 - Case series
  9 - Modeling
  0 - Other
"""

import csv
import os
import sqlite3
import sys

from itertools import groupby

import regex as re

from sklearn.ensemble import RandomForestClassifier

from .study import StudyModel
from .vocab import Vocab

class Design(StudyModel):
    """
    Prediction model used to classify study designs.
    """

    def __init__(self):
        """
        Builds a new Design model.

        Args:
            training: path to training data
            models: path to models
        """

        super(Design, self).__init__()

        # Keywords to use as features
        self.keywords = StudyModel.getKeywords()

    def predict(self, sections):
        # Build features array for document
        features = [self.features(sections)]

        return int(self.model.predict(features)[0])

    def create(self):
        return RandomForestClassifier(n_estimators=129, max_depth=21, max_features=0.22, random_state=0)

    def hyperparams(self):
        return {"n_estimators": range(125, 150),
                "max_depth": range(20, 30),
                "max_features": [x / 100 for x in range(15, 25)],
                "random_state": (0,)}

    def data(self, training):
        # Unique ids
        uids = {}

        # Features
        ids = []
        features = []
        labels = []

        # Unpack training data
        training, db = training
        cur = db.cursor()

        # Read training data, convert to features
        with open(training, mode="r") as csvfile:
            for row in csv.DictReader(csvfile):
                uids[row["id"]] = int(row["label"])

            # Build id list for each uid batch
            for idlist in self.batch(list(uids.keys()), 999):
                # Get section text and transform to features
                cur.execute("SELECT name, text, article FROM sections WHERE article in (%s) ORDER BY id" % ",".join(["?"] * len(idlist)), idlist)
                i, f, l = self.transform(cur.fetchall(), uids)

                # Combine lists from each batch
                ids.extend(i)
                features.extend(f)
                labels.extend(l)

        print("Loaded %d rows" % len(features))

        return ids, features, labels

    def batch(self, uids, size):
        """
        Splits uids into batches.

        Args:
            uids: uids
            size: batch size

        Returns:
            list of lists split into batch size
        """

        return [uids[x:x + size] for x in range(0, len(uids), size)]

    def transform(self, rows, uids):
        """
        Transforms a list of rows into features and labels.

        Args:
            rows: input rows
            uids: uid to label mapping

        Returns:
            (features, labels)
        """

        ids = []
        features = []
        labels = []

        # Retrieve all rows and group by article id
        for uid, sections in groupby(rows, lambda x: x[2]):
            # Get sections as list
            sections = list(sections)

            # Save ids, features and label
            ids.append(uid)
            features.append(self.features(sections))
            labels.append(uids[uid])

        return ids, features, labels

    def features(self, sections):
        """
        Builds a features vector from input text.

        Args:
            sections: list of sections

        Returns:
            features vector as a list
        """

        # Feature vector
        vector = []

        # Build title keyword features
        title = [text for name, text, _ in sections if name == "TITLE"]
        title = title[0].lower() if title else None

        for keyword in Vocab.TITLE:
            vector.append(len(re.findall("\\b%s\\b" % keyword.lower(), title)) if title else 0)

        # Build full text from filtered sections
        text = [text for name, text, _ in sections if not name or StudyModel.filter(name.lower())]
        text = " ".join(text).replace("\n", " ").lower()

        # Get token length across filtered sections
        length = sum([len(tokens) for name, _, tokens in sections if not name or StudyModel.filter(name.lower())])

        # Add study design term counts normalized by document length
        for keyword in self.keywords:
            vector.append(len(re.findall("\\b%s\\b" % keyword.lower(), text)) / length)

        return vector

    @staticmethod
    def run(training, path, optimize):
        """
        Trains a new model.

        Args:
            training: path to training file
            path: database path
            optimize: if hyperparameter optimization should be enabled
        """

        # Load articles database
        db = sqlite3.connect(path)

        try:
            # Train the model
            model = Design()
            model.train((training, db), optimize)

            # Save the model
            print("Saving model to %s" % path)
            model.save(os.path.join(path, "design"))

        finally:
            db.close()

if __name__ == "__main__":
    Design.run(sys.argv[1] if len(sys.argv) > 1 else None,
               sys.argv[2] if len(sys.argv) > 2 else None,
               sys.argv[3] == "1" if len(sys.argv) > 3 else False)



"""
Database module
"""

class Database(object):
    """
    Defines data structures and methods to store article content.
    """

    # pylint: disable=W0613
    def merge(self, url, ids):
        """
        Merges the results of an existing database into the current database. This method returns
        a list of ids not merged, which means there is a newer version available in the source data.

        Args:
            url: database connection
            ids: dict of id - entry date

        Returns:
            list of eligible ids NOT merged
        """

        return []

    def save(self, article):
        """
        Saves an article.

        Args:
            article: article metadata and text content
        """

    def complete(self):
        """
        Signals processing is complete and runs final storage methods.
        """

    def close(self):
        """
        Commits and closes the database.
        """



"""
PDF processing module
"""

from io import StringIO

import requests

from .tei import TEI

class PDF(object):
    """
    Methods to transform medical/scientific PDFs into article objects.
    """

    @staticmethod
    def parse(stream, source, models):
        """
        Parses a medical/scientific PDF datastream and returns a processed article.

        Args:
            stream: handle to input data stream
            source: text string describing stream source, can be None
            models: path to study models

        Returns:
            (uid, article metadata, section text, article tags, study design)
        """

        # Convert PDF stream to TEI XML, parse and return object
        return TEI.parse(PDF.convert(stream), source, models)

    @staticmethod
    def convert(stream):
        """
        Converts a medical/scientific article PDF into TEI XML via a GROBID Web Service API call.

        Args:
            stream: handle to input data stream

        Returns:
            TEI XML stream
        """

        # Call GROBID API
        data = requests.post("http://localhost:8070/api/processFulltextDocument", files={"input": stream}).text

        # Wrap as StringIO
        return StringIO(data)



"""
Vocabulary module

Credit to https://www.kaggle.com/savannareid for providing keywords and analysis.
"""

class Vocab(object):
    """
    Defines vocabulary terms for studies.
    """

    # Study titles
    TITLE = [r"case (?:series|study)", r"cross[\-\s]?sectional", r"mathematical model(?:ing)?", "meta-analysis",
             r"non[\-\s]randomized", "prospective cohort", "randomized", "retrospective cohort", "systematic review",
             r"time[\-\s]?series"]

    # Study design vocabulary
    DESIGN = [r"(?:electronic )?health records", r"(?:electronic )?medical records", "adherence", "adjusted hazard ratio",
              "adjusted odds ratio", "ahr", "aic", "akaike information criterion", "allocation method", "allocation method double-blind",
              "aor", "area under the curve", "associated with", "associated with random sample", "association", "attack rate", "auc"
              "baseline", "blind", "bootstrap", "bootstrap auc", "case report", "case report clinical findings", "case series",
              "case study", r"case[\-\s]control", "censoring", "chart review", "cochrane review", "coefficient", "cohen's d", "cohen's kappa",
              "cohort", r"computer model(?:ing)?", "confounding", "consort", "control arm", r"correlation(?:s)?",
              "covariates", "cox proportional hazards", "cross-sectional survey", r"cross[\-\s]sectional", "d-pooled", "data abstraction forms",
              "data collection instrument", r"database(?:s)? search(?:ed)?", "databases searched", r"deep[\-\s]learning", "demographics", "diagnosis",
              "difference between means", "difference in means", "dosage", "double-blind", "duration", "editor", "ehr", "electronic health records",
              "electronic search", "eligib(?:e|ility)", "eligibility criteria", r"enroll(?:ed|ment)?", "estimation", "etiology", "exclusion criteria",
              "exposure status", "follow-up", "followed", r"forecast(?:ing)?", "frequency", "gamma", "hazard ratio", "heterogeneity", "hr", "i2",
              "incidence", "inclusion criteria", "inter-rater reliability", "interrater reliability", "interventions", "kaplan-meier", "log odds",
              "logistic regression", "lognormal", "longitudinal", "loss to follow-up", r"machine[\-\s]learning", r"match(?:ed|ing)? case",
              r"match(?:ed|ing)? criteria", "matched", "matching", r"mathematical model(?:ing)?", "mean difference", "median time to event",
              "meta-analysis", "model fit", "model simulation", "monte carlo", "multivariate hazard ratio", "narrative review",
              "non-comparative study", r"non[\-\s]randomised", r"non[\-\s]randomized",
              "non-response bias", "number of controls per case", "odds", "odds ratio", "outcomes", "patients", "per capital", "placebo",
              "pooled adjusted odds ratio", "pooled aor", "pooled odds ratio", "pooled or", "pooled relative risk", "pooled risk ratio",
              "pooled rr", "potential confounders", "power", "prevalence", "prevalence survey", "prisma", "prospective cohort",
              r"prospective(?:ly)?", "protocol", "pseudo-randomised", "pseudo-randomized", "psychometric evaluation of instrument",
              "psychometric evaluaton of instrument", "publication bias", "quasi-randomised", "quasi-randomized", "questionnaire development",
              "r-squared", "randomisation", "randomisation consort", "randomised", "randomization method", "randomized", "randomized clinical trial",
              "randomized controlled trial", "rct", "receiver-operator curve", r"recruit(?:ed|ment)?", "registry", "registry data",
              "relative risk", "response rate", "retrospective", "retrospective chart review", "retrospective cohort", "right-censored",
              "risk factor analysis", "risk factors", "risk factors data collection instrument", "risk of bias", "risk ratio", "roc", "rr",
              "search criteria", "search strategy", "search string", r"simulat(?:e|ed|ion)", "statistical model", "stochastic model", "strength",
              "subjects", "surveillance", "survey instrument", "survival analysis", "symptoms", "syndromic surveillance", "synthetic",
              "synthetic data", r"synthetic data(?:set(?:s)?)?", "systematic review", "time-to-event analysis", r"time[\-\s]series",
              r"time[\-\s]varying", "tolerability", "treatment arm", "treatment effect", "truncated", "weibull"]

    # Sample vocabulary
    SAMPLE = ["articles", "cases", "children", "individuals", "men", "participants", "patients", "publications", "samples", "sequences",
              "studies", "trials", "total", "women"]

    # Sample methods vocabulary
    METHOD = ["analyse", "analyze", "ci", "clinical", "collect", "compare", "data", "database", "demographic", "enroll", "epidemiological",
              "evidence", "findings", "hospital", "include", "materials", r"method(?:s)?:?", "observe", "obtain", "perform", "publication",
              "publish", "recruit", r"result(?:s)?:?", "retrieve", "review", "search", "study", "studie"]



"""
Transforms and loads medical/scientific files into an articles database.
"""

import os

from ..factory import Factory
from .pdf import PDF
from .tei import TEI

class Execute(object):
    """
    Transforms and loads medical/scientific files into an articles database.
    """

    @staticmethod
    def run(indir, url, models):
        """
        Main execution method.

        Args:
            indir: input directory
            url: database url
            models: model directory
        """

        # Build database connection
        db = Factory.create(url)

        # Recursively walk directory looking for files
        for root, _, files in sorted(os.walk(indir)):
            for f in sorted(files):
                # Check if file ends with accepted extension
                if any([f.lower().endswith(ext) for ext in ["pdf", "xml"]]):
                    isPdf = f.lower().endswith("pdf")

                    # Build full path to file
                    path = os.path.join(root, f)

                    print("Processing: %s" % path)
                    with open(path, "rb" if isPdf else "r") as data:
                        # Parse and save record
                        db.save(PDF.parse(data, f, models) if isPdf else TEI.parse(data, f, models))

        # Complete and close database
        db.complete(None)
        db.close()



"""
Section module
"""

import json
import os
import re

from nltk.tokenize import sent_tokenize

from ..table import Table
from ..text import Text

class Section(object):
    """
    Parses text content from JSON file sections.
    """

    @staticmethod
    def parse(row, directory):
        """
        Reads title, abstract and body text for a given row. Text is returned as a list of sections.

        Args:
            row: input row
            directory: input directory

        Returns:
            list of text sections
        """

        sections = []

        # Add title and abstract sections
        for name in ["title", "abstract"]:
            text = row[name]
            if text:
                # Remove leading and trailing []
                text = re.sub(r"^\[", "", text)
                text = re.sub(r"\]$", "", text)

                # Transform and clean text
                text = Text.transform(text)

                sections.extend([(name.upper(), x) for x in sent_tokenize(text)])

        # Process each JSON file
        for path in Section.files(row):
            # Build article path
            article = os.path.join(directory, path)

            try:
                with open(article) as jfile:
                    data = json.load(jfile)

                    # Extract text from body
                    for section in data["body_text"]:
                        # Section name and text
                        name = section["section"].upper() if len(section["section"].strip()) > 0 else None
                        text = section["text"].replace("\n", " ")

                        # Clean and transform text
                        text = Text.transform(text)

                        # Split text into sentences, transform text and add to sections
                        sections.extend([(name, x) for x in sent_tokenize(text)])

                    # Extract text from tables
                    for name, entry in data["ref_entries"].items():
                        if "html" in entry and entry["html"]:
                            sections.extend([(name, x) for x in Table.parse(entry["html"])])

            # pylint: disable=W0703
            except Exception as ex:
                print("Error processing text file: {}".format(article), ex)

        # Filter out boilerplate elements from text
        return Section.filtered(sections)

    @staticmethod
    def files(row):
        """
        Build a list of json file paths to parse.

        Args:
            row: input row

        Returns:
            list of paths
        """

        paths = []

        # Build list of documents to parse
        for column in ["pdf_json_files", "pmc_json_files"]:
            if row[column]:
                paths.extend(row[column].split("; "))

        return paths

    @staticmethod
    def filtered(sections):
        """
        Returns a filtered list of text sections. Duplicate and boilerplate text strings are removed.

        Args:
            sections: input sections

        Returns:
            filtered list of sections
        """

        # Use list to preserve insertion order
        unique = []
        keys = set()

        # Boilerplate text to ignore
        boilerplate = ["COVID-19 resource centre", "permission to make all its COVID", "WHO COVID database",
                       "COVID-19 public health emergency response"]

        for name, text in sections:
            # Add unique text that isn't boilerplate text
            if not text in keys and not any([x in text for x in boilerplate]):
                unique.append((name, text))
                keys.add(text)

        return unique



"""
Sample module
"""

import numpy as np

from word2number import w2n

from .vocab import Vocab

class Sample(object):
    """
    Methods to extract the sample size of a study.
    """

    @staticmethod
    def extract(sections, attributes):
        """
        Attempts to extract sample size, sample and sample method using an attributes
        prediction model.

        Args:
            sections: list of sections
            attributes: attribute predictions for each section

        Returns:
            (size, sample, method)
        """

        size, sample, method = None, None, None

        # Get best attribute match per category
        best = np.argmax(attributes, axis=0)

        # labels - NO_MATCH, STATISTIC, SAMPLE_METHOD, SAMPLE_SIZE
        # section - name, text, tokens
        # Require minimum level of confidence for best prediction
        if best[2] >= 0.3:
            method = sections[best[2]][1]
        if best[3] >= 0.3:
            sample = sections[best[3]]

            # Attempt to extract size from sample size tokens
            # section - name, text, tokens
            size = Sample.find(sample[2], Vocab.SAMPLE)
            sample = sample[1]

        return size, sample, method

    @staticmethod
    def find(tokens, keywords):
        """
        Attempts to find a token that matches keywords having a numeric descriptor.
        i.e. 34 subjects, 30 patients, ten studies

        Args:
            token: NLP token
            keywords: list of keywords to match

        Returns:
            True if token matches keywords
        """

        matches = [Sample.match(token, keywords) for token in tokens]
        matches = [match for match in matches if match]

        return matches[0][0] if matches else None

    @staticmethod
    def match(token, keywords):
        """
        Compares a token against a list of keywords looking for a matching token with a
        numeric descriptor.

        Args:
            token: NLP token
            keywords: list of keywords to match

        Returns:
            list of matches if found or None
        """

        if token.text.lower() in keywords:
            return [Sample.tonumber(c.text) for c in token.children if Sample.isnumber(c)]

        return None

    @staticmethod
    def isnumber(token):
        """
        Determines if token represents a number.

        Args:
            token: input token

        Returns:
            True if this represents a number, False otherwise
        """

        # Returns true if following conditions are met:
        #  - Token POS is a number of it's all digits
        #  - Token DEP is in [amod, nummod]
        #  - None of the children are brackets (ignore citations [1], [2], etc)
        return (token.text.isdigit() or token.pos_ == "NUM") and token.dep_ in ["amod", "nummod"] and not any([c.text == "[" for c in token.children])

    @staticmethod
    def tonumber(token):
        """
        Attempts to convert a string to a number. Returns raw token if unsuccessful.

        Args:
            token: input token

        Returns:
            parsed token
        """

        try:
            return "%d" % w2n.word_to_num(token.replace(",", ""))
        # pylint: disable=W0702
        except:
            pass

        return token



"""
Attribute module
"""

import csv
import os
import sys

import numpy as np
import regex as re
import spacy

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression
from sklearn.multiclass import OneVsRestClassifier
from sklearn.preprocessing import MultiLabelBinarizer

from .sample import Sample
from .study import StudyModel
from .vocab import Vocab

class Attribute(StudyModel):
    """
    Prediction model used to classify study attributes such as the sample size, sampling method or risk factors.
    """

    def __init__(self):
        """
        Builds a new Attribute detection model.

        Args:
            training: path to training data
            models: path to models
        """

        super(Attribute, self).__init__()

        # Keywords to use as features
        self.keywords = StudyModel.getKeywords(design=False)

        # TF-IDF vectors
        self.tfidf = None

    def predict(self, sections):
        # Build features array for document
        text = [text for name, text, _ in sections if not name or StudyModel.filter(name.lower())]

        features = [self.features(text, tokens) for name, text, tokens in sections]

        # Build tf-idf vector
        vector = self.tfidf.transform([text for text, _ in features])

        # Concat tf-idf and features vector
        features = np.concatenate((vector.toarray(), [f for _, f in features]), axis=1)

        # Predict probability - run serially for each row to prevent multi-threaded GIL thrashing
        predictions = np.array([self.model.predict_proba([f])[0] for f in features])

        # Clear predictions for short text snippets and filtered sections
        for x, (name, text, _) in enumerate(sections):
            if len(text) <= 25 or (name and not StudyModel.filter(name.lower())):
                # Clear prediction
                # pylint: disable=E1136
                predictions[x] = np.zeros(predictions.shape[1])

        return predictions

    def create(self):
        return OneVsRestClassifier(LogisticRegression(C=0.995, solver="lbfgs", max_iter=1000, random_state=0))

    def hyperparams(self):
        return {"C": [x / 200 for x in range(100, 300)],
                "solver": ("lbfgs", "liblinear"),
                "max_iter": (1000,),
                "random_state": (0,)}

    def data(self, training):
        # Features
        ids = []
        features = []
        labels = []

        # Load NLP model to parse tokens
        nlp = spacy.load("en_core_sci_md")

        # Read training data, convert to features
        with open(training, mode="r") as csvfile:
            for row in csv.DictReader(csvfile):
                # Parse text tokens
                text = row["text"]
                tokens = nlp(text)

                # Store ids, features and labels
                ids.append(text)
                features.append(self.features(text, tokens))
                labels.append([int(label) for label in row["label"].split(":")])

        # Build tf-idf model across dataset, concat with feature vector
        self.tfidf = TfidfVectorizer()
        vector = self.tfidf.fit_transform([text for text, _ in features])
        features = np.concatenate((vector.toarray(), [f for _, f in features]), axis=1)

        print("Loaded %d rows" % len(features))

        labels = MultiLabelBinarizer().fit_transform(labels)

        return ids, features, labels

    def features(self, text, tokens):
        """
        Builds a features vector from input text.

        Args:
            sections: list of sections

        Returns:
            features vector as a list
        """

        # Feature vector
        vector = []

        # Add study design term counts normalized by number of tokens
        for keyword in self.keywords:
            vector.append(len(re.findall("\\b%s\\b" % keyword.lower(), text.lower())) / len(tokens))

        pos = [token.pos_ for token in tokens]
        dep = [token.dep_ for token in tokens]

        # Append entity count (scispacy only tracks generic entities) normalized by number of tokens
        vector.append(len([entity for entity in tokens.ents if entity.text.lower() in self.keywords]) / len(tokens))

        # Append part of speech counts normalized by number of tokens
        for name in ["ADJ", "ADP", "ADV", "AUX", "CONJ", "CCONJ", "DET", "INTJ", "NOUN", "NUM", "PART", "PRON", "PUNCT",
                     "SCONJ", "SYM", "VERB", "X", "SPACE"]:
            vector.append(pos.count(name) / len(tokens))

        # Append dependency counts normalized by number of tokens
        for name in ["acl", "advcl", "advmod", "amod", "appos", "aux", "case", "cc", "ccomp", "clf", "compound",
                     "conj", "cop", "csubj", "dep", "det", "discourse", "dislocated", "expl", "fixed", "flat",
                     "goeswith", "iobj", "list", "mark", "nmod", "nsubj", "nummod", "obj", "obl", "orphan",
                     "parataxis", "punct", "reparandum", "root", "vocative", "xcomp"]:
            vector.append(dep.count(name) / len(tokens))

        # Descriptive numbers on sample identifiers - i.e. 34 patients, 15 subjects, ten samples
        vector.append(1 if Sample.find(tokens, Vocab.SAMPLE) else 0)

        # Regular expression for dates
        dateregex = r"(January|Jan|February|Feb|March|Mar|April|Apr|May|June|Jun|July|Jul|August|Aug|" + \
                    r"September|Sep|October|Oct|November|Nov|December|Dec)\s?\d{1,2}?,? \d{4}?"

        # Dates within the string normalized by number of tokens
        dates = len(re.findall(dateregex, text))
        vector.append(dates / len(tokens))

        return (text, vector)

    @staticmethod
    def run(training, path, optimize):
        """
        Trains a new model.

        Args:
            training: path to training file
            path: models path
            optimize: if hyperparameter optimization should be enabled
        """

        # Train the model
        model = Attribute()
        model.train(training, optimize)

        # Save the model
        print("Saving model to %s" % path)
        model.save(os.path.join(path, "attribute"))

if __name__ == "__main__":
    Attribute.run(sys.argv[1] if len(sys.argv) > 1 else None,
                  sys.argv[2] if len(sys.argv) > 2 else None,
                  sys.argv[3] == "1" if len(sys.argv) > 3 else False)



"""
Table module
"""

import re

from lxml import etree

class Table(object):
    """
    Parses text content from HTML tables.
    """

    @staticmethod
    def parse(text):
        """
        Parses content from a HTML table string. Builds a list of header-value pairs for each row.

        Args:
            text: HTML table string

        Returns:
            list of header-value pairs for each row
        """

        # Parse HTML content using lxml
        # pylint: disable=c-extension-no-member
        table = etree.HTML(text).find("body/table")

        return Table.extract(table)

    @staticmethod
    def extract(table):
        """
        Parses content from a HTML table element. Builds a list of header-value pairs for each row.

        Args:
            table: HTML table element

        Returns:
            list of header-value pairs for each row
        """

        # Table rows
        output = []

        if len(table) > 0:
            rows = iter(table)
            headers = [col.text for col in next(rows)]

            for row in rows:
                # Build concatenated header value string
                values = ["%s %s" % (headers[x] if x < len(headers) else "", column.text) for x, column in enumerate(row)]

                # Create single row string
                value = " ".join(values)

                # Remove whitespace
                value = re.sub(r"[\n\xa0\t]|\s{2,}", " ", value).strip()
                if value:
                    output.append(value)

        return output



"""
Grammar module
"""

import en_core_sci_md

class Grammar(object):
    """
    Linguistic processing rules applied to sentences.
    """

    def __init__(self):
        # spaCy NLP
        self.nlp = en_core_sci_md.load()

    def parse(self, text):
        """
        Parses list of text elements to NLP tokens.

        Args:
            text: list of text elements

        Returns:
            tokenslist - list of token lists
        """

        results = None

        if text:
            # Run text through linguistic rules
            results = [self.applyRules(tokens) for tokens in self.nlp.pipe(text, batch_size=256)]

        return results

    def label(self, tokens):
        """
        Linguistic rules processing logic. Identifies non-informative sentences and labels them accordingly.

        Labels:
            - QUESTION: Any text ending in a question mark
            - FRAGMENT: Poorly structured sentences with limited information

        Args:
            tokens: parsed tokens

        Returns:
            label if detected else None
        """

        label = None

        if tokens:
            # Label non-informative sentences
            if self.isQuestion(tokens.text):
                label = "QUESTION"
            elif self.isFragment(tokens):
                label = "FRAGMENT"

        return label

    def applyRules(self, tokens):
        """
        Apply custom rules to the parsed tokens.

        Args:
            tokens: parsed tokens

        Returns:
            updated tokens with rules applied
        """

        # Override POS to NOUN for 2019-nCoV
        for token in tokens:
            if token.pos_ == "NUM" and token.text.lower() == "2019-ncov":
                token.pos_ = "NOUN"

        return tokens

    def isQuestion(self, text):
        """
        Determines if the text is a question

        Args:
            text: input text

        Returns:
            true if text is a question, false otherwise
        """

        # Questions have a ? mark at end
        return text.strip().endswith("?")

    def isFragment(self, tokens):
        """
        Run text against linguistic rules to determine if sentence is a fragment. Fragments are non descriptive.

        Args:
            tokens: nlp document tokens

        Returns:
            true if text is a sentence fragment, false otherwise
        """

        # Nominal subject Nouns/Proper Nouns
        nouns = any([t.pos_  in ["NOUN", "PROPN"]  and t.dep_ in ["nsubj", "nsubjpass"] for t in tokens])

        # Actions (adverb, auxiliary, verb)
        action = any([t.pos_ in ["ADV", "AUX", "VERB"] for t in tokens])

        # Consider root words with a nominal subject as an action word
        action = action or any([t.dep_ in ["appos", "ROOT"] and any([x.dep_ in ["nsubj", "nsubjpass"] for x in t.children]) for t in tokens])

        # Non-punctuation tokens and multi-character words (don't count single letters which are often variables used in equations)
        words = [t.text for t in tokens if t.pos_ not in ["PUNCT", "SPACE", "SYM"] and len(t.text) > 1]

        # Valid sentences take the following form:
        #  - At least one nominal subject noun/proper noun AND
        #  - At least one action/verb AND
        #  - At least 5 words
        valid = nouns and action and len(words) >= 5

        return not valid



"""
Defines main entry point for ETL process.
"""

import sys

from .execute import Execute

if __name__ == "__main__":
    if len(sys.argv) > 1:
        Execute.run(sys.argv[1],
                    sys.argv[2] if len(sys.argv) > 2 else None,
                    sys.argv[3] if len(sys.argv) > 3 else None,
                    sys.argv[4] if len(sys.argv) > 4 else None,
                    sys.argv[5] == "True" if len(sys.argv) > 5 else True,
                    sys.argv[6] if len(sys.argv) > 6 else None)



"""
FileSystem module
"""

import json
import os

import yaml

from .database import Database

class FileSystem(Database):
    """
    Defines data structures and methods to store articles in a filesystem directory.
    """

    def __init__(self, outdir):
        """
        Creates and initializes an output directory to use for writing files.

        Args:
            outdir: output directory
        """

        # Create if output path doesn't exist
        os.makedirs(outdir, exist_ok=True)

        self.outdir = outdir

    def save(self, article):
        output = os.path.splitext(article.source if article.source else article.uid())[0] + ".%s" % self.extension()

        with open(os.path.join(self.outdir, output), "w") as output:
            self.write(output, article.build())

    def extension(self):
        """
        Returns file extension for generated files
        """

    def write(self, output, article):
        """
        Writes article content to output stream.

        Args:
            output: output file handle
            article: article object
        """

class JSON(FileSystem):
    """
    Defines data structures and methods to store articles as JSON files.
    """

    def extension(self):
        return "json"

    def write(self, output, article):
        json.dump(article, output)

class YAML(FileSystem):
    """
    Defines data structures and methods to store articles as YAML files.
    """

    def extension(self):
        return "yaml"

    def write(self, output, article):
        output.write(yaml.safe_dump(article))



"""
SQLite module
"""

import os
import sqlite3

from datetime import datetime, timedelta

from .database import Database

class SQLite(Database):
    """
    Defines data structures and methods to store article content in SQLite.
    """

    # Articles schema
    ARTICLES = {
        "Id": "TEXT PRIMARY KEY",
        "Source": "TEXT",
        "Published": "DATETIME",
        "Publication": "TEXT",
        "Authors": "TEXT",
        "Title": "TEXT",
        "Tags": "TEXT",
        "Design": "INTEGER",
        "Size": "TEXT",
        "Sample": "TEXT",
        "Method": "TEXT",
        "Reference": "TEXT",
        "Entry": "DATETIME"
    }

    # Sections schema
    SECTIONS = {
        "Id": "INTEGER PRIMARY KEY",
        "Article": "TEXT",
        "Tags": "TEXT",
        "Design": "INTEGER",
        "Name": "TEXT",
        "Text": "TEXT",
        "Labels": "TEXT"
    }

    # SQL statements
    CREATE_TABLE = "CREATE TABLE IF NOT EXISTS {table} ({fields})"
    INSERT_ROW = "INSERT INTO {table} ({columns}) VALUES ({values})"
    CREATE_INDEX = "CREATE INDEX section_article ON sections(article)"

    # Merge SQL statements
    ATTACH_DB = "ATTACH DATABASE '{path}' as {name}"
    DETACH_DB = "DETACH DATABASE '{name}'"
    MAX_ENTRY = "SELECT MAX(entry) from {name}.articles"
    LOOKUP_ARTICLE = "SELECT Id FROM {name}.articles WHERE Id=? AND Entry = ?"
    MERGE_ARTICLE = "INSERT INTO articles SELECT * FROM {name}.articles WHERE Id = ?"
    MERGE_SECTIONS = "INSERT INTO sections SELECT * FROM {name}.sections WHERE Article=?"
    UPDATE_ENTRY = "UPDATE articles SET entry = ? WHERE Id = ?"
    ARTICLE_COUNT = "SELECT COUNT(1) FROM articles"
    SECTION_COUNT = "SELECT MAX(id) FROM sections"

    def __init__(self, outdir):
        """
        Creates and initializes a new output SQLite database.

        Args:
            outdir: output directory
        """

        # Create if output path doesn't exist
        os.makedirs(outdir, exist_ok=True)

        # Output database file
        dbfile = os.path.join(outdir, "articles.sqlite")

        # Delete existing file
        if os.path.exists(dbfile):
            os.remove(dbfile)

        # Index fields
        self.aindex, self.sindex = 0, 0

        # Create output database
        self.db = sqlite3.connect(dbfile)

        # Create database cursor
        self.cur = self.db.cursor()

        # Create articles table
        self.create(SQLite.ARTICLES, "articles")

        # Create sections table
        self.create(SQLite.SECTIONS, "sections")

        # Start transaction
        self.cur.execute("BEGIN")

    def merge(self, url, ids):
        # List of IDs to set for processing
        queue = set()

        # Attached database alias
        alias = "merge"

        # Attach database
        self.db.execute(SQLite.ATTACH_DB.format(path=url, name=alias))

        # Only process records newer than 5 days before the last run
        lastrun = self.cur.execute(SQLite.MAX_ENTRY.format(name=alias)).fetchone()[0]
        lastrun = datetime.strptime(lastrun, "%Y-%m-%d") - timedelta(days=5)
        lastrun = lastrun.strftime("%Y-%m-%d")

        # Search for existing articles
        for uid, date in ids.items():
            self.cur.execute(SQLite.LOOKUP_ARTICLE.format(name=alias), [uid, date])
            if not self.cur.fetchone() and date > lastrun:
                # Add uid to process
                queue.add(uid)
            else:
                # Copy existing record
                self.cur.execute(SQLite.MERGE_ARTICLE.format(name=alias), [uid])
                self.cur.execute(SQLite.MERGE_SECTIONS.format(name=alias), [uid])

                # Sync entry date with ids list
                self.cur.execute(SQLite.UPDATE_ENTRY, [date, uid])

        # Set current index positions
        self.aindex = int(self.cur.execute(SQLite.ARTICLE_COUNT.format(name=alias)).fetchone()[0]) + 1
        self.sindex = int(self.cur.execute(SQLite.SECTION_COUNT.format(name=alias)).fetchone()[0]) + 1

        # Commit transaction
        self.db.commit()

        # Detach database
        self.db.execute(SQLite.DETACH_DB.format(name=alias))

        # Start new transaction
        self.cur.execute("BEGIN")

        # Return list of new/updated ids to process
        return queue

    def save(self, article):
        # Article row
        self.insert(SQLite.ARTICLES, "articles", article.metadata)

        # Increment number of articles processed
        self.aindex += 1
        if self.aindex % 1000 == 0:
            print("Inserted {} articles".format(self.aindex), end="\r")

            # Commit current transaction and start a new one
            self.transaction()

        for name, text, labels in article.sections:
            # Section row - id, article, tags, design, name, text, labels
            self.insert(SQLite.SECTIONS, "sections", (self.sindex, article.uid(), article.tags(), article.design(), name, text, labels))
            self.sindex += 1

    def complete(self):
        print("Total articles inserted: {}".format(self.aindex))

        # Create articles index for sections table
        self.execute(SQLite.CREATE_INDEX)

    def close(self):
        self.db.commit()
        self.db.close()

    def transaction(self):
        """
        Commits current transaction and creates a new one.
        """

        self.db.commit()
        self.cur.execute("BEGIN")

    def create(self, table, name):
        """
        Creates a SQLite table.

        Args:
            table: table schema
            name: table name
        """

        columns = ["{0} {1}".format(name, ctype) for name, ctype in table.items()]
        create = SQLite.CREATE_TABLE.format(table=name, fields=", ".join(columns))

        # pylint: disable=W0703
        try:
            self.cur.execute(create)
        except Exception as e:
            print(create)
            print("Failed to create table: " + e)

    def execute(self, sql):
        """
        Executes SQL statement against open cursor.

        Args:
            sql: SQL statement
        """

        self.cur.execute(sql)

    def insert(self, table, name, row):
        """
        Builds and inserts a row.

        Args:
            table: table object
            name: table name
            row: row to insert
        """

        # Build insert prepared statement
        columns = [name for name, _ in table.items()]
        insert = SQLite.INSERT_ROW.format(table=name,
                                          columns=", ".join(columns),
                                          values=("?, " * len(columns))[:-2])

        try:
            # Execute insert statement
            self.cur.execute(insert, self.values(table, row, columns))
        # pylint: disable=W0703
        except Exception as ex:
            print("Error inserting row: {}".format(row), ex)

    def values(self, table, row, columns):
        """
        Formats and converts row into database types based on table schema.

        Args:
            table: table schema
            row: row tuple
            columns: column names

        Returns:
            Database schema formatted row tuple
        """

        values = []
        for x, column in enumerate(columns):
            # Get value
            value = row[x]

            if table[column].startswith("INTEGER"):
                values.append(int(value) if value else 0)
            elif table[column].startswith("BOOLEAN"):
                values.append(1 if value == "TRUE" else 0)
            elif table[column].startswith("TEXT"):
                # Clean empty text and replace with None
                values.append(value if value and len(value.strip()) > 0 else None)
            else:
                values.append(value)

        return values



"""
Study module
"""

import math
import pickle

import numpy as np
import regex as re

from sklearn.metrics import accuracy_score, log_loss, f1_score
from sklearn.model_selection import train_test_split, GridSearchCV

from .vocab import Vocab

class StudyModel(object):
    """
    Base StudyModel class used for study classifiers.
    """

    SECTION_FILTER = r"background|(?<!.*?results.*?)discussion|introduction|reference"

    @staticmethod
    def getKeywords(design=True, sample=True, method=True):
        """
        Generates an unique keyword vector list.

        Returns:
            list of keyword regular expressions
        """

        keywords = []
        if design:
            keywords = keywords + Vocab.DESIGN
        if sample:
            keywords = keywords + Vocab.SAMPLE
        if method:
            keywords = keywords + Vocab.METHOD

        # Deduplicate keywords
        return sorted(set(keywords))

    @staticmethod
    def filter(name):
        """
        Filters a section name. Returns True if name is a title, method or results section.

        Args:
            name: section name

        Returns:
            True if section should be analyzed, False otherwise
        """

        # Skip background, introduction and reference sections
        # Skip discussion unless it's a results and discussion
        return not re.search(StudyModel.SECTION_FILTER, name)

    def __init__(self):
        """
        Builds a new StudyModel.
        """

        # Prediction model
        self.model = None
        self.tfidf = None

        # Keywords
        self.keywords = None

    def load(self, path):
        """
        Loads a StudyModel.

        Args:
            path: path to model
        """

        with open(path, "rb") as handle:
            self.__dict__ = pickle.load(handle)

    def save(self, path):
        """
        Saves a StudyModel.

        Args:
            path: path to model
        """

        with open(path, "wb") as handle:
            pickle.dump(self.__dict__, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def train(self, training, optimize):
        """
        Trains a new model.

        Args:
            training: training data
            optimize: if hyperparameters should be optimized
        """

        # Create model, parameters set via hyperparameter tuning
        model = self.create()

        # Enable hyperparameter optimization for this run
        if optimize:
            params = self.hyperparams()
            model = GridSearchCV(model, params, cv=5, verbose=1, n_jobs=-1)

        # Load training data
        ids, features, labels = self.data(training)

        # Split into train/test
        x_train, x_test, y_train, y_test = train_test_split(features, labels, test_size=0.1, random_state=0)

        # Fit the model
        model.fit(x_train, y_train)

        # Reporting
        if optimize:
            print("Optimal Parameters - ", model.best_params_)

        # Store model
        self.model = model

        # Score model
        self.score(x_test, y_test)

        # Test model predictions on ALL training data
        # Helps show records model is confused about within training data
        for i, x in enumerate(features):
            pred = model.predict([x])
            if not np.array_equal(pred[0], labels[i]):
                print(ids[i], labels[i], ", WRONG: ", pred)

    def score(self, features, labels):
        """
        Scores the model accuracy for test data.

        Args:
            features: test features
            labels: test labels
        """

        predictions = self.model.predict_proba(features)
        size = range(len(predictions[0]))
        multiclass = False

        # Handle both single-class and multi-class labels
        if isinstance(labels, np.ndarray):
            plabels = [[x >= 0.5 for x in p] for p in predictions]
            multiclass = True
        else:
            plabels = [p.argmax() for p in predictions]

        print("Test Accuracy: ", accuracy_score(labels, plabels))
        print("Test F1 Score: ", f1_score(labels, plabels, labels=size, average="weighted", zero_division=0))
        print("Test Log Loss: ", log_loss(labels, predictions, labels=size if not multiclass else None))
        print("Base Log Loss (nclass=%d): " % len(size), -math.log(1 / len(size)))

    def predict(self, sections):
        """
        Classifies sections using a prediction model.

        Args:
            sections: list of text sections

        Returns:
            (label, probability array)
        """

    def create(self):
        """
        Creates a new model.

        Returns:
            (model, tfidf)
        """

        return (None, None)

    def hyperparams(self):
        """
        Returns a range of hyperparameters.

        Returns:
            dict of hyperparameter ranges
        """

        return {}

    def data(self, training):
        """
        Reads training data.

        Args:
            training: training input file

        Returns:
            (ids, features, labels)
        """

        return (training, training, training)



# pylint: disable = C0111
from setuptools import find_packages, setup

with open("README.md", "r") as f:
    DESCRIPTION = f.read()

setup(name="paperetl",
      version="1.2.0",
      author="NeuML",
      description="ETL processes for medical and scientific papers",
      long_description=DESCRIPTION,
      long_description_content_type="text/markdown",
      url="https://github.com/neuml/paperetl",
      project_urls={
          "Documentation": "https://github.com/neuml/paperetl",
          "Issue Tracker": "https://github.com/neuml/paperetl/issues",
          "Source Code": "https://github.com/neuml/paperetl",
      },
      license="Apache 2.0: http://www.apache.org/licenses/LICENSE-2.0",
      packages=find_packages(where="src/python"),
      package_dir={"": "src/python"},
      keywords="etl parse covid-19 medical scientific papers",
      python_requires=">=3.6",
      install_requires=[
          "beautifulsoup4>=4.8.1",
          "elasticsearch>=7.8.0",
          "lxml>=4.5.0",
          "nltk>=3.5",
          "numpy>=1.18.4",
          "python-dateutil>=2.8.1",
          "PyYAML>=5.3",
          "regex>=2020.5.14",
          "requests>=2.22.0",
          "scikit-learn>=0.23.1",
          "scipy>=1.4.1",
          "spacy>=2.3.2",
          "word2number>=1.1"
      ],
      classifiers=[
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Topic :: Software Development",
          "Topic :: Utilities"
      ])



"""
Analysis module
"""

import os

import numpy as np

from .study.attribute import Attribute
from .study.design import Design
from .study.sample import Sample

# Global helper for multi-processing support
# pylint: disable=W0603
MODELS = None

def getModels(models):
    """
    Multiprocessing helper method. Gets (or first creates then gets) a global study models object to
    be accessed in a new subprocess.

    Args:
        models: models directory

    Returns:
        (attribute model, design model)
    """

    global MODELS

    if not MODELS:
        attribute = Attribute()
        attribute.load(os.path.join(models, "attribute"))

        design = Design()
        design.load(os.path.join(models, "design"))

        MODELS = (attribute, design)

    return MODELS

class Study(object):
    """
    Study design parser. Derives study design fields using a series of machine learning models and NLP analysis.
    """

    @staticmethod
    def parse(sections, models):
        """
        Parses study design fields contained within an article.

        Args:
            sections: list of text sections
            models: models directory

        Returns:
            study design fields as tuple
        """

        # Get design models
        attribute, design = getModels(models)

        # Study design type
        design = design.predict(sections)

        # Detect attributes within sections
        attributes = attribute.predict(sections)

        # Extract sample size, sample, method
        size, sample, method = Sample.extract(sections, attributes)

        # Label each section
        labels = []
        for x, (_, text, _) in enumerate(sections):
            # Get predicted attribute
            attribute = np.argmax(attributes[x])
            label = None

            if attribute == 1:
                label = "STATISTIC"
            elif text == sample:
                label = "SAMPLE_SIZE"
            elif text == method:
                label = "SAMPLE_METHOD"

            labels.append(label)

        return (design, size, sample, method, labels)



"""
Defines main entry point for ETL process.
"""

import sys

from .execute import Execute

if __name__ == "__main__":
    if len(sys.argv) > 3:
        Execute.run(sys.argv[1], sys.argv[2], sys.argv[3])



"""
Article module.
"""

class Article(object):
    """
    Article objects. Holds all metadata and text content related to an article.
    """

    # Articles schema
    ARTICLE = ("id", "source", "published", "publication", "authors", "title", "tags", "design",
               "size", "sample", "method", "reference", "entry")

    # Sections schema
    SECTION = ("name", "text", "labels")

    def __init__(self, metadata, sections, source):
        """
        Stores article metadata and section content as an object.

        Args:
            metadata: article metadata
            sections: text sections
            source: article source
        """

        self.metadata = metadata
        self.sections = sections
        self.source = source

    def uid(self):
        """
        Returns the article uid.

        Returns:
            article uid
        """

        return self.metadata[0]

    def tags(self):
        """
        Returns the article tags.

        Returns:
            article tags
        """

        return self.metadata[6]

    def design(self):
        """
        Returns the article design.

        Returns:
            article design
        """

        return self.metadata[7]

    def build(self):
        """
        Builds an article with all metadata and section content.

        Returns:
            object with full article content
        """

        # Create article
        article = dict(zip(Article.ARTICLE, self.metadata))

        # Create sections
        sections = [dict(zip(Article.SECTION, section)) for section in self.sections]

        # Add sections to article
        article["sections"] = sections

        return article



"""
Transforms and loads CORD-19 data into an articles database.
"""

import csv
import hashlib
import os.path
import re

from datetime import datetime
from multiprocessing import Pool
from dateutil import parser

from ..analysis import Study
from ..factory import Factory
from ..grammar import Grammar
from ..schema.article import Article
from .section import Section

# Global helper for multi-processing support
# pylint: disable=W0603
GRAMMAR = None

def getGrammar():
    """
    Multiprocessing helper method. Gets (or first creates then gets) a global grammar object to
    be accessed in a new subprocess.

    Returns:
        Grammar
    """

    global GRAMMAR

    if not GRAMMAR:
        GRAMMAR = Grammar()

    return GRAMMAR

class Execute(object):
    """
    Transforms and loads CORD-19 data into an articles database.
    """

    @staticmethod
    def getHash(row):
        """
        Gets sha hash for this row. Builds one from the title if no body content is available.

        Args:
            row: input row

        Returns:
            sha1 hash id
        """

        # Use sha1 provided, if available
        sha = row["sha"].split("; ")[0] if row["sha"] else None
        if not sha:
            # Fallback to sha1 of title
            sha = hashlib.sha1(row["title"].encode("utf-8")).hexdigest()

        return sha

    @staticmethod
    def getDate(row):
        """
        Parses the publish date from the input row.

        Args:
            row: input row

        Returns:
            publish date
        """

        date = row["publish_time"]

        if date:
            try:
                if date.isdigit() and len(date) == 4:
                    # Default entries with just year to Jan 1
                    date += "-01-01"

                return parser.parse(date)

            # pylint: disable=W0702
            except:
                # Skip parsing errors
                return None

        return None

    @staticmethod
    def getUrl(row):
        """
        Parses the url from the input row.

        Args:
            row: input row

        Returns:
            url
        """

        if row["url"]:
            # Filter out API reference links
            urls = [url for url in row["url"].split("; ") if "https://api." not in url]
            if urls:
                return urls[0]

        # Default to DOI
        return "https://doi.org/"  + row["doi"]

    @staticmethod
    def getTags(sections):
        """
        Searches input sections for matching keywords. If found, returns the keyword tag.

        Args:
            sections: list of text sections

        Returns:
            tags
        """

        # Keyword patterns to search for
        keywords = [r"2019[\-\s]?n[\-\s]?cov", "2019 novel coronavirus", "coronavirus 2(?:019)?", r"coronavirus disease (?:20)?19",
                    r"covid(?:[\-\s]?(?:20)?19)?", r"n\s?cov[\-\s]?2019", r"sars[\-\s]cov-?2", r"wuhan (?:coronavirus|cov|pneumonia)"]

        # Build regular expression for each keyword. Wrap term in word boundaries
        regex = "|".join(["\\b%s\\b" % keyword.lower() for keyword in keywords])

        tags = None
        for _, text in sections:
            # Look for at least one keyword match
            if re.findall(regex, text.lower()):
                tags = "COVID-19"
                break

        return tags

    @staticmethod
    def stream(indir, models, dates, merge):
        """
        Generator that yields rows from a metadata.csv file. The directory is also included.

        Args:
            indir: input directory
            models: models directory
            dates: list of uid - entry dates for current metadata file
            merge: only merges/processes this list of uids, if enabled
        """

        # Filter out duplicate ids
        ids, hashes = set(), set()

        with open(os.path.join(indir, "metadata.csv"), mode="r") as csvfile:
            for row in csv.DictReader(csvfile):
                # cord uid
                uid = row["cord_uid"]

                # sha hash
                sha = Execute.getHash(row)

                # Only process if all conditions below met:
                #  - Merge mode false or uid in list of ids to merge
                #  - cord uid in entry date mapping
                #  - cord uid and sha hash not already processed
                if (not merge or uid in merge) and uid in dates and uid not in ids and sha not in hashes:
                    yield (row, indir, models)

                # Add uid and sha as processed
                ids.add(uid)
                hashes.add(sha)

    @staticmethod
    def process(params):
        """
        Processes a single row

        Args:
            params: (row, indir, models)

        Returns:
            (id, article, sections)
        """

        # Get grammar handle
        grammar = getGrammar()

        # Unpack parameters
        row, indir, models = params

        # Published date
        date = Execute.getDate(row)

        # Get text sections
        sections = Section.parse(row, indir)

        # Search recent documents for COVID-19 keywords
        tags = Execute.getTags(sections) if not date or date >= datetime(2019, 7, 1) else None

        if tags:
            # Build NLP tokens for sections
            tokenslist = grammar.parse([text for _, text in sections])

            # Join NLP tokens with sections
            sections = [(name, text, tokenslist[x]) for x, (name, text) in enumerate(sections)]

            # Parse study design fields
            design, size, sample, method, labels = Study.parse(sections, models)

            # Add additional fields to each section
            sections = [(name, text, labels[x] if labels[x] else grammar.label(tokens)) for x, (name, text, tokens) in enumerate(sections)]
        else:
            # Untagged section, create None default placeholders
            design, size, sample, method = None, None, None, None

            # Extend sections with empty columns
            sections = [(name, text, None) for name, text in sections]

        # Article metadata - id, source, published, publication, authors, title, tags, design, sample size
        #                    sample section, sample method, reference
        metadata = (row["cord_uid"], row["source_x"], date, row["journal"], row["authors"], row["title"], tags, design, size,
                    sample, method, Execute.getUrl(row))

        return Article(metadata, sections, None)

    @staticmethod
    def entryDates(indir, entryfile):
        """
        Loads an entry date lookup file into memory.

        Args:
            indir: input directory
            entryfile: path to entry dates file

        Returns:
            dict of cord uid -> entry date
        """

        # sha - (cord uid, date) mapping
        entries = {}

        # Default path to entry files if not provided
        if not entryfile:
            entryfile = os.path.join(indir, "entry-dates.csv")

        # Load in memory date lookup
        with open(entryfile, mode="r") as csvfile:
            for row in csv.DictReader(csvfile):
                entries[row["sha"]] = (row["cord_uid"], row["date"])

        # Reduce down to entries only in metadata
        dates = {}
        with open(os.path.join(indir, "metadata.csv"), mode="r") as csvfile:
            for row in csv.DictReader(csvfile):
                # Lookup hash
                sha = Execute.getHash(row)

                # Lookup record
                uid, date = entries[sha]

                # Store date if cord uid maps to value in entries
                if row["cord_uid"] == uid:
                    dates[uid] = date

        return dates

    @staticmethod
    def run(indir, url, models, entryfile, full, merge):
        """
        Main execution method.

        Args:
            indir: input directory
            url: database url
            models: model directory
            entryfile: path to entry dates file
            full: full database load if True, only loads tagged articles if False
            merge: database url to use for merging prior results
        """

        print("Building articles database from {}".format(indir))

        # Set database url and models defaults if not provided
        if not url:
            url = os.path.join(os.path.expanduser("~"), ".cord19", "models")
            models = url

        # Create database
        db = Factory.create(url)

        # Load entry dates
        dates = Execute.entryDates(indir, entryfile)

        # Merge existing db, if present
        if merge:
            merge = db.merge(merge, dates)
            print("Merged results from existing articles database")

        # Create process pool
        with Pool(os.cpu_count()) as pool:
            for article in pool.imap(Execute.process, Execute.stream(indir, models, dates, merge), 100):
                # Get unique id
                uid = article.uid()

                # Only load untagged rows if this is a full database load
                if full or article.tags():
                    # Append entry date
                    article.metadata = article.metadata + (dates[uid],)

                    # Save article
                    db.save(article)

        # Complete processing
        db.complete()

        # Commit and close
        db.close()
